#  tests for twisted-22.2.0-py311h80987f9_1 (this is a generated file);
print('===== testing package: twisted-22.2.0-py311h80987f9_1 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
import twisted
import os

print('twisted.__version__: %s' % twisted.__version__)
assert twisted.__version__ == os.getenv("PKG_VERSION")
#  --- run_test.py (end) ---

print('===== twisted-22.2.0-py311h80987f9_1 OK =====');
print("import: 'twisted'")
import twisted

print("import: 'twisted._threads'")
import twisted._threads

print("import: 'twisted.application'")
import twisted.application

print("import: 'twisted.application.runner'")
import twisted.application.runner

print("import: 'twisted.application.twist'")
import twisted.application.twist

print("import: 'twisted.conch'")
import twisted.conch

print("import: 'twisted.conch.client'")
import twisted.conch.client

print("import: 'twisted.conch.insults'")
import twisted.conch.insults

print("import: 'twisted.conch.openssh_compat'")
import twisted.conch.openssh_compat

print("import: 'twisted.conch.scripts'")
import twisted.conch.scripts

print("import: 'twisted.conch.ssh'")
import twisted.conch.ssh

print("import: 'twisted.conch.ui'")
import twisted.conch.ui

print("import: 'twisted.cred'")
import twisted.cred

print("import: 'twisted.enterprise'")
import twisted.enterprise

print("import: 'twisted.internet'")
import twisted.internet

print("import: 'twisted.logger'")
import twisted.logger

print("import: 'twisted.mail'")
import twisted.mail

print("import: 'twisted.mail.scripts'")
import twisted.mail.scripts

print("import: 'twisted.names'")
import twisted.names

print("import: 'twisted.pair'")
import twisted.pair

print("import: 'twisted.persisted'")
import twisted.persisted

print("import: 'twisted.plugins'")
import twisted.plugins

print("import: 'twisted.positioning'")
import twisted.positioning

print("import: 'twisted.protocols'")
import twisted.protocols

print("import: 'twisted.protocols.haproxy'")
import twisted.protocols.haproxy

print("import: 'twisted.python'")
import twisted.python

print("import: 'twisted.runner'")
import twisted.runner

print("import: 'twisted.scripts'")
import twisted.scripts

print("import: 'twisted.spread'")
import twisted.spread

print("import: 'twisted.tap'")
import twisted.tap

print("import: 'twisted.trial'")
import twisted.trial

print("import: 'twisted.trial._dist'")
import twisted.trial._dist

print("import: 'twisted.web'")
import twisted.web

print("import: 'twisted.web._auth'")
import twisted.web._auth

print("import: 'twisted.words'")
import twisted.words

print("import: 'twisted.words.im'")
import twisted.words.im

print("import: 'twisted.words.protocols'")
import twisted.words.protocols

print("import: 'twisted.words.protocols.jabber'")
import twisted.words.protocols.jabber

print("import: 'twisted.words.xish'")
import twisted.words.xish

